/// Statistics of hash distributed search.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct HdSearchStatistics {
    /// Number of nodes expanded by each thread.
    pub expanded: Vec<usize>,
    /// Number of states generated by each thread.
    pub generated: Vec<usize>,
    /// Number of nodes kept by each thread.
    pub kept: Vec<usize>,
    /// Number of nodes sent to other threads by each thread.
    pub sent: Vec<usize>,
}
